package de.hftstuttgart.unifiedticketing.core;

import de.hftstuttgart.unifiedticketing.exceptions.AssertionException;

import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * <b>Ticket Builder</b><br/>
 * <br/>
 * This class provides a builder mechanism, to define a new Ticket and create it afterwards.
 * @param <B> implementation of this class this instance belongs to
 * @param <T> type of ticket implementation this builder will create
 * @param <TS> type of ticket system this builder belongs to
 */
public abstract class TicketBuilder<B extends TicketBuilder, T extends Ticket, TS extends TicketSystem>
{
    private static Logger logger = Logging.getLogger(TicketBuilder.class.getName());

    public final TS parent;

    protected Set<String> assignees;
    protected String description;
    protected Set<String> labels;
    protected String title;

    protected TicketBuilder(TS parent)
    {
        if (parent == null)
        {
            logger.log(Level.SEVERE, "every ticket builder instance needs a belonging ticket system, but received null!");
            throw new AssertionException(
                String.format("Got null as parent ticket system on creating new %s", this.getClass().getSimpleName()));
        }
        this.parent = parent;
        logger.log(Level.FINEST, String.format("%s instantiated", this.getClass().getSimpleName()));
    }

    public B assignees(String... identifiers)
    {
        this.assignees = new HashSet<>(Arrays.asList(identifiers));
        return (B) this;
    }

    public B description(String description)
    {
        this.description = description;
        logger.log(Level.FINEST, "description added to new Ticket");
        return (B) this;
    }

    public B labels(Set<String> labels)
    {
        this.labels = labels;
        logger.log(Level.FINEST, "labels added to new Ticket");
        return (B) this;
    }

    public B labels(String[] labels)
    {
        logger.log(Level.FINEST, "transforming labels Array to Set");
        return this.labels(new HashSet<>(Arrays.asList(labels)));
    }

    public B title(String title)
    {
        this.title = title;
        logger.log(Level.FINEST, "title added to new Ticket");
        return (B) this;
    }

    /**
     * persists the defined ticket into a create request to the ticket system connected
     * @return newly created ticket
     */
    public abstract T create();
}
