package de.hftstuttgart.unifiedticketing.systems.gitlab;

import de.hftstuttgart.unifiedticketing.core.Logging;
import de.hftstuttgart.unifiedticketing.core.TicketSystem;
import de.hftstuttgart.unifiedticketing.exceptions.AssertionException;
import de.hftstuttgart.unifiedticketing.exceptions.UnifiedticketingException;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class GitlabTicketSystem extends TicketSystem<GitlabTicket, GitlabTicketSystem, GitlabTicketBuilder, GitlabFilter>
{
    private static Logger logger = Logging.getLogger(TicketSystem.class.getName());

    /**
     * creates a new instance of this class from an uri.<br/>
     * <br/>
     * <b>Attention:</b> This method should not be called from the enduser!<br/>
     * Instead call the same method from {@link TicketSystem}.<br/>
     * To call this method directly, the uri has to be shortend by the "unifiedticketing:gitlab:" part.
     * @param uri gitlab specific part of full uri
     * @return new gitlab ticket instance
     */
    public static GitlabTicketSystem fromUri(String uri)
    {
        Matcher matcher = Pattern.compile("^((http|https):\\/\\/)?(.*(:[0-9]+.*)?):([0-9]+)(:(.*))?$").matcher(uri);

        if (!matcher.matches())
        {
            logger.log(Level.SEVERE, "uri didn't match regex");
            throw new AssertionException("uri didn't match regex");
        }

        GitlabTicketSystemBuilder builder = new GitlabTicketSystemBuilder();
        if (matcher.group(2) != null && matcher.group(2).length() > 0)
        {
            if (matcher.group(2).equalsIgnoreCase("http")) builder.withHttp();
            else builder.withHttps();
        }

        if (matcher.group(3) == null || matcher.group(3).length() == 0)
        {
            logger.log(Level.SEVERE, "no base url identified in uri");
            throw new AssertionException("no base url found");
        }
        builder.withBaseUrl(matcher.group(3));

        if (matcher.group(5) == null || matcher.group(5).length() == 0)
        {
            logger.log(Level.SEVERE, "no project id identified in uri");
            throw new AssertionException("no project id found");
        }
        builder.withProjectId(matcher.group(5));

        if (matcher.group(6) == null || matcher.group(6).length() == 0)
        {
            logger.log(Level.INFO, "no api key identified in uri, creating anonymous instance");
        } else
        {
            builder.withApiKey(matcher.group(7));
        }

        return builder.build();
    }

    protected GitlabTicketSystem()
    {
        logger.log(Level.FINEST,"GitLab Ticketsystem created");
    }

    /**
     * starts builder process for a new gitlab ticket
     */
    @Override
    public GitlabTicketBuilder createTicket()
    {
        return new GitlabTicketBuilder(this);
    }

    /**
     * starts filter builder to define constraints and then request those tickets
     */
    @Override
    public GitlabFilter find()
    {
        return new GitlabFilter(this);
    }

//    @Override
//    public List<GitlabTicket> getAllTickets()
//    {
//        logger.log(Level.FINER, "redirecting request to find method");
//        return this.find()
//            .get();
//    }

    /**
     * This requests a dedicated ticket by it's id.
     * It internally calls {@link #find()}, set's the id as filter
     * and ensure to get exactly one element as answer.
     * @param id ticket id to request
     * @throws UnifiedticketingException if more or less than one element in the response
     */
    @Override
    public GitlabTicket getTicketById(String id)
    {
        logger.log(Level.FINER, "redirecting request to find method");
        List<GitlabTicket> ret = this.find()
            .withId(id)
            .get();

        if (ret == null) return null;
        else if (ret.size() != 1) throw new UnifiedticketingException("more or less than 1 dedicated ticket found");
        else return ret.get(0);
    }

    @Override
    public boolean hasAssigneeSupport() { return true; }

    @Override
    public boolean hasDefaultPagination() { return true; }

    @Override
    public boolean hasLabelSupport() { return true; }

    @Override
    public boolean hasPaginationSupport() { return true; }

    @Override
    public boolean hasReturnNullOnErrorSupport() { return true; }
}
