package de.hftstuttgart.unifiedticketing.systems.gitlab;

import com.fasterxml.jackson.databind.ObjectMapper;
import de.hftstuttgart.unifiedticketing.exceptions.AssertionException;
import de.hftstuttgart.unifiedticketing.exceptions.DeserializationException;
import de.hftstuttgart.unifiedticketing.exceptions.HttpResponseException;
import okhttp3.*;
import okio.Buffer;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mockito;

import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

public class GitlabTicketTest
{
    public GitlabTicket instance;
    public Call call;
    public ArgumentCaptor<Request> requestCaptor;
    public Response.Builder responseBuilder;

    @BeforeEach
    public void initBeforeEach()
    {
        instance = spy(new GitlabTicket(mock(GitlabTicketSystem.class)));
        instance.getParent().baseUrl = "https://example.org";

        OkHttpClient client = mock(OkHttpClient.class);
        call = mock(Call.class);
        requestCaptor = ArgumentCaptor.forClass(Request.class);
        responseBuilder = new Response.Builder()
            .request(new Request.Builder().url("http://test.some.tld/").build())
            .protocol(Protocol.HTTP_1_1)
            .message("some message");

        doReturn(client).when(instance).getHttpClient();
        doReturn(call).when(client).newCall(requestCaptor.capture());
    }

    @Test
    public void testFromTicketResponse()
    {
        GitlabTicketResponse res = new GitlabTicketResponse();
        GitlabTicket ticket = GitlabTicket.fromTicketResponse(instance.getParent(), res);

        assertAll(
            () -> assertSame(instance.getParent(), ticket.getParent()),
            () -> assertNull(ticket.getTitle()),
            () -> assertNull(ticket.getDescription()),
            () -> assertEquals("0", ticket.getId()),
            () -> assertNotNull(ticket.getLabels()),
            () -> assertTrue(ticket.isOpen()),
            () -> assertNull(ticket.getTitle()),
            () -> assertNotNull(ticket.getAssignees())
        );
    }

    @Test
    public void testAddAssigneeByString()
    {
        assertThrows(AssertionException.class, () -> instance.addAssignee("blbal"));

        instance = Mockito.spy(instance);
        int id = 4;
        instance.addAssignee(String.valueOf(id));
        verify(instance, times(1)).addAssignee(id);
    }

    @Test
    public void testAddAssigneeByInt()
    {
        int id = 5;
        instance.addAssignee(id);
        assertEquals(1, instance.getAssignees().size());
        assertEquals(String.valueOf(id), instance.getAssignees().stream().findAny().get().id);

        for (int i=0; i < 5; i++)
        {
            instance.addAssignee(id);
        }
        assertEquals(1, instance.getAssignees().size());
    }

    @Test
    public void testRemoveAssigneeByString()
    {
        assertThrows(AssertionException.class, () -> instance.removeAssignee("nonparsableint"));

        instance = Mockito.spy(instance);
        int id = 4;
        instance.removeAssignee(String.valueOf(id));
        verify(instance, times(1)).removeAssignee(id);
    }

    @Test
    public void testRemoveAssigneeByInt()
    {
        int[] ids = new int[]{4, 8, 19, 500};
        for (int id: ids)
        {
            instance.addAssignee(id);
        }

        assertEquals(ids.length, instance.getAssignees().size());
        instance.removeAssignee(ids[0]);
        assertEquals(ids.length -1, instance.getAssignees().size());
        Set<Integer> leftover = instance.getAssignees().stream()
            .map(a -> Integer.parseInt(a.id))
            .collect(Collectors.toSet());

        for (int i=1; i < ids.length; i++)
        {
            assertTrue(leftover.contains(ids[i]));
        }
    }

    @Test
    public void testSaveNoHttpClientWithoutChanges()
    {
        GitlabTicket actual = instance.save();
        verify(instance, never()).getHttpClient();
        assertSame(instance, actual);
    }

    @Test
    public void testSaveServerError() throws IOException
    {
        doReturn(responseBuilder.code(500).build()).when(call).execute();
        instance.setTitle("we have a new title");
        assertThrows(HttpResponseException.class, () -> instance.save());
    }

    @Test
    public void testSaveClientError() throws IOException
    {
        doReturn(responseBuilder.code(400).build()).when(call).execute();
        instance.setTitle("we have another new title");
        assertThrows(HttpResponseException.class, () -> instance.save());
    }

    @Test
    public void testSaveNullBody() throws IOException
    {
        doReturn(responseBuilder.code(200).build()).when(call).execute();
        instance.setTitle("we have no body this time");
        assertThrows(HttpResponseException.class, () -> instance.save());
    }

    @Test
    public void testSaveNoJsonBody() throws IOException
    {
        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create("somestrangething", MediaType.get("application/json")))
                .build())
            .when(call).execute();
        instance.setTitle("this body is no json");
        assertThrows(DeserializationException.class, () -> instance.save());
    }

    @Test
    public void testSaveSuccessfulUpdate() throws IOException
    {
        GitlabTicketResponse ticketResponse = new GitlabTicketResponse();
        ticketResponse.title = "title of ticket";
        ticketResponse.description = "description";
        ticketResponse.iid = 5;
        ticketResponse.labels = new HashSet<>(Arrays.asList("bug", "unifiedticketing"));
        GitlabTicketResponse.Assignee assignee = new GitlabTicketResponse.Assignee();
        assignee.id = 234;
        assignee.name = "Full name";
        assignee.username = "username";
        ticketResponse.assignees = new LinkedList<>(Collections.singleton(assignee));

        GitlabTicket expected = GitlabTicket.fromTicketResponse(instance.getParent(), ticketResponse);
        String responseJson = new ObjectMapper().writeValueAsString(ticketResponse);

        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create(responseJson.getBytes(), MediaType.get("application/json")))
                .build())
            .when(call).execute();
        GitlabTicket actual = instance.close().save();
        assertEquals(expected, actual);
        assertTrue(expected.deepEquals(actual));
    }

    @Test
    public void testSaveRequestJson() throws IOException
    {
        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create("{}".getBytes(), MediaType.get("application/json")))
                .build())
            .when(call).execute();

        instance.setTitle("title of ticket")
            .setDescription("description")
            .setLabels(new HashSet<>(Arrays.asList("bug", "unifiedticketing")))
            .addAssignee(234)
            .save();

        Buffer buffer = new Buffer();
        ObjectMapper mapper = new ObjectMapper();
        requestCaptor.getValue().body().writeTo(buffer);
        String expectedJson = "{\"assignee_ids\":[234],\"description\":\"description\"," +
            "\"title\":\"title of ticket\",\"labels\":\"bug,unifiedticketing\"}";
        assertEquals(mapper.readTree(expectedJson), mapper.readTree(buffer.readUtf8()));
    }
}
